// (C) Copyright Raffi Enficiaud 2017.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at 
// http://www.boost.org/LICENSE_1_0.txt)

// See http://www.boost.org/libs/test for the library home page.
//
//! @file
//! Forwarding source
// **************************************************************************

#define BOOST_TEST_SOURCE
#include <boost/test/impl/test_framework_init_observer.ipp>

// EOF
